import cv2
import numpy as np
import sys
from moviepy.video.io.VideoFileClip import VideoFileClip
import wave
import contextlib
import os
import struct
from tqdm import tqdm
import matplotlib.pyplot as plt
import math
from collections import Counter as CounterDict
from Crypto.Cipher import AES
from Crypto.Util import Counter as AESCounter
from Crypto.Random import get_random_bytes

# ===================
# Parametry AVRNG
# ===================
FRAME_SKIP = 0        # Zignoruj pierwsze sekundy
VT_SCALE = 0.3        # Współczynnik progowy różnicy RGB
THRESHOLD = 80       # Limit prób znalezienia innego punktu
K = 7000              # Liczba bajtów dźwięku do miksowania
BITS_PER_BYTE = 8

# ===================
# Funkcja pomocnicza
# ===================
def aes256_ctr_postprocess(bits):
    
    # Przekształć bity na bajty
    byte_array = bytearray([
        int("".join(bits[i:i+8]), 2)
        for i in range(0, len(bits) - 7, 8)
    ])

    # Wygeneruj losowy klucz AES-256 i licznik
    key = get_random_bytes(32)  # 256-bit key
    ctr = AESCounter.new(128)

    cipher = AES.new(key, AES.MODE_CTR, counter=ctr)
    encrypted_bytes = cipher.encrypt(byte_array)

    # Zamień zaszyfrowane bajty z powrotem na bity
    encrypted_bits = []
    for byte in encrypted_bytes:
        bin_str = format(byte, "08b")
        encrypted_bits.extend(list(bin_str))

    return encrypted_bits

def cml_postprocess(raw_bits, epsilon=0.6, N=1000, k=26):
    
    assert len(raw_bits) >= 256, "Wejściowy ciąg bitów musi mieć ≥ 256 bitów"
    processed_bits = []

    for i in range(0, len(raw_bits) - 255, 256):
        segment = raw_bits[i:i + 256]

        # Podziel segment na 8 × 32 bity i zamień na float (0, 1)
        states = []
        for j in range(8):
            bits32 = segment[j * 32:(j + 1) * 32]
            int_val = int("".join(bits32), 2)
            float_val = int_val / (2**32 - 1)  # normalizacja do [0,1]
            states.append(float_val)

        # Funkcja logistyczna
        def f(x): return 4 * x * (1 - x)

        # Iteracje mapy CML
        states = np.array(states, dtype=np.float64)
        for _ in range(N):
            new_states = np.zeros_like(states)
            for j in range(8):
                left = states[(j - 1) % 8]
                center = states[j]
                right = states[(j + 1) % 8]
                new_states[j] = (1 - epsilon) * f(center) + (epsilon / 2) * (f(left) + f(right))
            states = new_states

        # Zamień końcowe stany na k-bitowe liczby binarne
        for x in states:
            x = max(0.0, min(x, 1.0))  # upewnij się że x ∈ [0,1]
            int_kbit = int(x * (2**k - 1))
            bin_kbit = format(int_kbit, f'0{k}b')
            processed_bits.extend(list(bin_kbit))

    return processed_bits


def calculate_byte_entropy(bits):
    # Grupuj bity w bajty
    byte_values = [
        int("".join(bits[i:i + 8]), 2)
        for i in range(0, len(bits) - 7, 8)
    ]

    # Zlicz częstość bajtów
    counter = CounterDict(byte_values)
    total = len(byte_values)

    entropy = 0.0
    for count in counter.values():
        p = count / total
        entropy -= p * math.log2(p)

    print(f"Entropia bajtów: {entropy:.6f}")

def plot_byte_histogram(bits, output_file="byte_histogram.png"):
    # Grupowanie bitów w bajty
    byte_values = [
        int("".join(bits[i:i + 8]), 2)
        for i in range(0, len(bits) - 7, 8)
    ]

    # Histogram
    plt.figure(figsize=(8, 4))
    plt.hist(byte_values, bins=range(257), density=True, color="black")
    plt.title("Empiryczny rozkład zmiennych losowych generowanych przez źródło szumu")
    plt.xlabel("Wartość (x)")
    plt.ylabel("Częstość występowania $P_x$")
    plt.grid(True, axis='y', linestyle=":", linewidth=0.5)
    plt.tight_layout()
    plt.savefig(output_file, dpi=150)
    plt.show()

def get_initial_coordinate(frame):
    h, w, _ = frame.shape
    x_c, y_c = w // 2, h // 2
    patch = frame[y_c-1:y_c+2, x_c-1:x_c+2]
    color_i = np.mean(patch, axis=(0,1)).astype(np.uint8)
    color_code = (int(color_i[2]) << 16) + (int(color_i[1]) << 8) + int(color_i[0])
    x = (color_code % (w // 2)) + w // 4
    y = (color_code % (h // 2)) + h // 4
    return int(x), int(y)

def extract_audio_bytes(video_path):
    clip = VideoFileClip(video_path)
    audio_path = "temp_audio.wav"
    clip.audio.write_audiofile(audio_path, fps=44100, nbytes=2, buffersize=2000, codec='pcm_s16le')
    with contextlib.closing(wave.open(audio_path, 'rb')) as wf:
        frames = wf.readframes(wf.getnframes())
        audio_bytes = np.frombuffer(frames, dtype=np.int16)
    os.remove(audio_path)
    return audio_bytes

def avrng(video_path, n_bits_required, output_path="random_output.txt"):
    cap = cv2.VideoCapture(video_path)
    audio_bytes = extract_audio_bytes(video_path)
    frame_rate = int(cap.get(cv2.CAP_PROP_FPS))
    width  = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

    vt = None
    r1 = g1 = b1 = r2 = g2 = b2 = 0
    x, y = 0, 0
    watchdog = 0
    runcnt = 0
    bits = []

    frame_count = 0
    sound_index = 0

    print("Generowanie bitów...")

    while len(bits) < n_bits_required and cap.isOpened():
        ret, frame = cap.read()
        if not ret:
            cap.set(cv2.CAP_PROP_POS_FRAMES, 0)
            continue


        frame_count += 1
        if frame_count < FRAME_SKIP * frame_rate:
            continue

        if vt is None:
            vt = VT_SCALE * np.var(frame)

        h, w, _ = frame.shape
        x, y = get_initial_coordinate(frame)

        for _ in range(w * h):
            R, G, B = map(int, frame[y, x])
            diff = (int(R) - r1) ** 2 + (int(G) - g1) ** 2 + (int(B) - b1) ** 2

            if diff < vt:
                x = (x + (R ^ G) + 1) % w
                y = (y + (G ^ B) + 1) % h
                watchdog += 1
                if watchdog > THRESHOLD:
                    break
                continue

            sn = []
            for offset, shift in zip([10, 15, 20, 5, 25], [0, 2, 3, 1, 5]):
                index = offset + (R * runcnt + (G << shift) + B + runcnt) % (K // 2)
                safe_index = int((sound_index + index) % len(audio_bytes))
                sn_val = int(audio_bytes[safe_index])
                sn.append(sn_val)

            bit = 1 & (R ^ G ^ B ^ r1 ^ g1 ^ b1 ^ r2 ^ g2 ^ b2 ^ sn[0] ^ sn[1] ^ sn[2] ^ sn[3] ^ sn[4])
            bits.append(str(bit))

            if len(bits) % 100000 == 0:
                runcnt += 1

            r1, g1, b1 = R, G, B
            x = (((R ^ x) << 4) ^ (G ^ y)) % w
            y = (((G ^ x) << 4) ^ (B ^ y)) % h
            r2, g2, b2 = R, G, B
            sound_index += K

            if len(bits) >= n_bits_required:
                break
            if len(bits) % 8000 == 0:
                x, y = get_initial_coordinate(frame)

    cap.release()

    usable_bits = bits[:len(bits) // 256 * 256]

    print("Wykonywanie postprocessingu CML-MPRBG...")
    processed_bits = cml_postprocess(usable_bits)

    print("Postprocessing AES-256-CTR...")
    encrypted_bits = aes256_ctr_postprocess(processed_bits)

    with open(output_path, 'w') as f:
        f.write("".join(bits))

    print(f"Zapisano {len(encrypted_bits)} bitów do pliku: {output_path}")
    plot_byte_histogram(encrypted_bits)
    calculate_byte_entropy(encrypted_bits)

# ===================
# Uruchomienie
# ===================
if __name__ == "__main__":
    if len(sys.argv) < 3:
        print("Użycie: python trng.py <ścieżka_do_pliku_video> <liczba_bitów>")
        sys.exit(1)

    video_path = sys.argv[1]
    n_bits = int(sys.argv[2])
    avrng(video_path, n_bits)
